// lexertl.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include "extras/cpp.h"
#include "include/lexertl/debug.hpp"
#include "include/lexertl/generator.hpp"
#include "include/lexertl/iterator.hpp"
#include "include/lexertl/memory_file.hpp"
#include "extras/rev_regex.h"

#include "include/lexertl/generate_cpp.hpp"

#include <iomanip>
#include <iostream>
#include <regex>

const char *names_ [] = { "T_ERROR",
	"T_AND_ALT",
	"T_ANDASSIGN_ALT",
	"T_ANDAND_ALT",
	"T_OR_ALT",
	"T_ORASSIGN_ALT",
	"T_OROR_ALT",
	"T_XORASSIGN_ALT",
	"T_XOR_ALT",
	"T_NOTEQUAL_ALT",
	"T_NOT_ALT",
	"T_COMPL_ALT",
	"T_IMPORT",
	"T_ARROWSTAR",
	"T_DOTSTAR",
	"T_COLON_COLON",
	"T_AND",
	"T_ANDAND",
	"T_ASSIGN",
	"T_ANDASSIGN",
	"T_OR",
	"T_OR_TRIGRAPH",
	"T_ORASSIGN",
	"T_ORASSIGN_TRIGRAPH",
	"T_XOR",
	"T_XOR_TRIGRAPH",
	"T_XORASSIGN",
	"T_XORASSIGN_TRIGRAPH",
	"T_COMMA",
	"T_COLON",
	"T_DIVIDEASSIGN",
	"T_DIVIDE",
	"T_DOT",
	"T_ELLIPSIS",
	"T_EQUAL",
	"T_GREATER",
	"T_GREATEREQUAL",
	"T_LEFTBRACE",
	"T_LEFTBRACE_ALT",
	"T_LEFTBRACE_TRIGRAPH",
	"T_LESS",
	"T_LESSEQUAL",
	"T_LEFTPAREN",
	"T_LEFTBRACKET",
	"T_LEFTBRACKET_ALT",
	"T_LEFTBRACKET_TRIGRAPH",
	"T_MINUS",
	"T_MINUSASSIGN",
	"T_MINUSMINUS",
	"T_PERCENT",
	"T_PERCENTASSIGN",
	"T_NOT",
	"T_NOTEQUAL",
	"T_OROR",
	"T_OROR_TRIGRAPH",
	"T_PLUS",
	"T_PLUSASSIGN",
	"T_PLUSPLUS",
	"T_ARROW",
	"T_QUESTION_MARK",
	"T_RIGHTBRACE",
	"T_RIGHTBRACE_ALT",
	"T_RIGHTBRACE_TRIGRAPH",
	"T_RIGHTPAREN",
	"T_RIGHTBRACKET",
	"T_RIGHTBRACKET_ALT",
	"T_RIGHTBRACKET_TRIGRAPH",
	"T_SEMICOLON",
	"T_SHIFTLEFT",
	"T_SHIFTLEFTASSIGN",
	"T_SHIFTRIGHT",
	"T_SHIFTRIGHTASSIGN",
	"T_STAR",
	"T_COMPL",
	"T_COMPL_TRIGRAPH",
	"T_STARASSIGN",
	"T_ASM",
	"T_AUTO",
	"T_BOOL",
	"T_FALSE",
	"T_TRUE",
	"T_BREAK",
	"T_CASE",
	"T_CATCH",
	"T_CHAR",
	"T_CLASS",
	"T_CONST",
	"T_CONSTCAST",
	"T_CONTINUE",
	"T_DEFAULT",
	"T_DELETE",
	"T_DO",
	"T_DOUBLE",
	"T_DYNAMICCAST",
	"T_ELSE",
	"T_ENUM",
	"T_EXPLICIT",
	"T_EXPORT",
	"T_EXTERN",
	"T_FLOAT",
	"T_FOR",
	"T_FRIEND",
	"T_GOTO",
	"T_IF",
	"T_INLINE",
	"T_INT",
	"T_LONG",
	"T_MUTABLE",
	"T_NAMESPACE",
	"T_NEW",
	"T_OPERATOR",
	"T_PRIVATE",
	"T_PROTECTED",
	"T_PUBLIC",
	"T_REGISTER",
	"T_REINTERPRETCAST",
	"T_RETURN",
	"T_SHORT",
	"T_SIGNED",
	"T_SIZEOF",
	"T_STATIC",
	"T_STATICCAST",
	"T_STRUCT",
	"T_SWITCH",
	"T_TEMPLATE",
	"T_THIS",
	"T_THROW",
	"T_TRY",
	"T_TYPEDEF",
	"T_TYPEID",
	"T_TYPENAME",
	"T_UNION",
	"T_UNSIGNED",
	"T_USING",
	"T_VIRTUAL",
	"T_VOID",
	"T_VOLATILE",
	"T_WCHART",
	"T_WHILE",
	"T_PP_DEFINE",
	"T_PP_IF",
	"T_PP_IFDEF",
	"T_PP_IFNDEF",
	"T_PP_ELSE",
	"T_PP_ELIF",
	"T_PP_ENDIF",
	"T_PP_ERROR",
	"T_PP_QHEADER",
	"T_PP_HHEADER",
	"T_PP_INCLUDE",
	"T_PP_LINE",
	"T_PP_PRAGMA",
	"T_PP_UNDEF",
	"T_PP_WARNING",
	"T_MSEXT_INT8",
	"T_MSEXT_INT16",
	"T_MSEXT_INT32",
	"T_MSEXT_INT64",
	"T_MSEXT_BASED",
	"T_MSEXT_DECLSPEC",
	"T_MSEXT_CDECL",
	"T_MSEXT_FASTCALL",
	"T_MSEXT_STDCALL",
	"T_MSEXT_TRY",
	"T_MSEXT_EXCEPT",
	"T_MSEXT_FINALLY",
	"T_MSEXT_LEAVE",
	"T_MSEXT_INLINE",
	"T_MSEXT_ASM",
	"T_MSEXT_PP_REGION",
	"T_MSEXT_PP_ENDREGION",
	"T_LONGINTLIT",
	"T_INTLIT",
	"T_FLOATLIT",
	"T_IDENTIFIER",
	"T_CCOMMENT",
	"T_CPPCOMMENT",
	"T_CHARLIT",
	"T_STRINGLIT",
	"T_SPACE",
	"T_SPACE2",
	"T_CONTLINE",
	"T_NEWLINE",
	"T_POUND_POUND",
	"T_POUND_POUND_ALT",
	"T_POUND_POUND_TRIGRAPH",
	"T_POUND",
	"T_POUND_ALT",
	"T_POUND_TRIGRAPH",
	"T_ANY_TRIGRAPH",
	"T_ANY",
	"T_ALIGNAS",
	"T_ALIGNOF",
	"T_CHAR16_T",
	"T_CHAR32_T",
	"T_CONSTEXPR",
	"T_DECLTYPE",
	"T_NOEXCEPT",
	"T_NULLPTR",
	"T_STATICASSERT",
	"T_THREADLOCAL",
	"T_RAWSTRINGLIT"};
/*
void rev_lex()
{
	lexertl::rules rules;
	lexertl::state_machine sm;

	build_cpp(rules);
	rules.reverse();
	lexertl::generator::build(rules, sm);

	lexertl::memory_file mf_("include/lexertl/enums.hpp");
	typedef std::reverse_iterator<const char *> rev_iter;
	typedef lexertl::match_results<rev_iter> match;
	typedef lexertl::iterator<rev_iter, lexertl::state_machine, match> iter;
	//		lexertl::citerator iter_(mf_.data(), mf_.data() + mf_.size(), sm);
	//		lexertl::citerator end_;
	iter iter_(rev_iter(mf_.data() + mf_.size()), rev_iter(mf_.data()), sm);
	iter end_;

	for (; iter_ != end_; ++iter_)
	{
		if (iter_->id == iter_->npos())
		{
			throw std::runtime_error("WAT");
		}
		else
		{
			std::cout << names_[iter_->id] << ", " << iter_->str() << '\n';
		}
	}
}
*/
void macro_tests()
{
	/*
	rules.insert_macro("any", "[\t\v\f\r\n\040-\377]");
	rules.insert_macro("Backslash", "[\\\\]|\"??/\"");
	rules.insert_macro("HexDigit", "[a-fA-F0-9]");
	rules.insert_macro("OctalDigit", "[0-7]");
	rules.insert_macro("HexQuad", "{HexDigit}{HexDigit}{HexDigit}{HexDigit}");
	rules.insert_macro("EscapeSequence",
	"{Backslash}([abfnrtv?'\"]|{Backslash}|x{HexDigit}+|"
	"{OctalDigit}{OctalDigit}?{OctalDigit}?)");
	rules.insert_macro("UniversalChar", "{Backslash}(u{HexQuad}|U{HexQuad}{HexQuad})");

	rules.push("L?([\"]({EscapeSequence}|{UniversalChar}|"
	"{any}{-}[\n\r\\\\\"])*[\"])", T_STRINGLIT);
	lexertl::generator::build(rules, sm);
	lexertl::debug::dump(sm, std::cout);

	std::string input_("\"stdafx.h\"");

	lexertl::citerator iter_(input_.c_str(), input_.c_str() + input_.size(), sm);
	lexertl::citerator end_;

	for (; iter_ != end_; ++iter_)
	{
	}
	*/

	/*
	rules.insert_macro("AllValid", "[\\x1-\\xff]");
	rules.insert_macro("ControlCodes", "[\\0-\\x1f\\x7f-\\x9f]");
	rules.insert_macro("AnyCh", "{AllValid}{-}{ControlCodes}");
	rules.push("{AnyCh}+", 123);

	lexertl::generator::build(rules, sm);
	lexertl::debug::dump(sm, std::cout);
	*/
}
/*
void openscad()
{
	lexertl::rules rules;
	enum {eEOF, eInclude, eFilePath, eFileName, eUseFileName,
		TOK_USE, TOK_MODULE, TOK_FUNCTION, TOK_IF, TOK_ELSE, TOK_LET,
		TOK_FOR, TOK_TRUE, TOK_FALSE, TOK_UNDEF, TOK_ERROR, TOK_NUMBER,
		TOK_ID, eNewLine, eTab, eReturn, eBackslash, eDblQuote, eError,
		eHex, eUnicode, eChar, TOK_STRING, LE, GE, EQ, NE, AND, OR, eDot};

	rules.push_state("cond_comment");
	rules.push_state("cond_lcomment");
	rules.push_state("cond_string");
	rules.push_state("cond_include");
	rules.push_state("cond_use");

	rules.insert_macro("D", "[0-9]");
	rules.insert_macro("E", "[Ee][+-]?{D}+");
	rules.insert_macro("H", "[0-9a-fA-F]");

	rules.insert_macro("U", "[\x80-\xbf]");
	rules.insert_macro("U2", "[\xc2-\xdf]");
	rules.insert_macro("U3", "[\xe0-\xef]");
	rules.insert_macro("U4", "[\xf0-\xf4]");
	rules.insert_macro("UNICODE", "{U2}{U}|{U3}{U}{U}|{U4}{U}{U}{U}");

	rules.push("INITIAL", "include[ \t\r\n>]*<", rules.skip(), "cond_include");
	rules.push("cond_include", "[^\t\r\n>]*[/]", eFilePath, ".");
	rules.push("cond_include", "[^\t\r\n>/]+", eFileName, ".");
	rules.push("cond_include", ">", eInclude, "INITIAL");

	rules.push("INITIAL", "use[ \t\r\n>]*<", rules.skip(), "cond_use");
	rules.push("cond_use", "[^\t\r\n>]+", eUseFileName, ".");
	rules.push("cond_use", ">", TOK_USE, "INITIAL");

	rules.push("module", TOK_MODULE);
	rules.push("function", TOK_FUNCTION);
	rules.push("if", TOK_IF);
	rules.push("else", TOK_ELSE);
	rules.push("let",  TOK_LET);
	rules.push("for", TOK_FOR);

	rules.push("true", TOK_TRUE);
	rules.push("false",	TOK_FALSE);
	rules.push("undef",	TOK_UNDEF);

	rules.push("(\xc2\xa0)+", rules.skip());

	rules.push("{UNICODE}+", TOK_ERROR);

	rules.push("{D}+{E}?|{D}*[.]{D}+{E}?|{D}+[.]{D}*{E}?", TOK_NUMBER);
	rules.push("[$]?[a-zA-Z0-9_]+", TOK_ID);

	rules.push("INITIAL", "[\"]", rules.skip(), "cond_string");
	rules.push("cond_string", "\n", eNewLine, ".");
	rules.push("cond_string", "\t", eTab, ".");
	rules.push("cond_string", "\r", eReturn, ".");
	rules.push("cond_string", "\\\\", eBackslash, ".");
	rules.push("cond_string", "\\\"", eDblQuote, ".");
	rules.push("cond_string", "{UNICODE}", eError, ".");
	rules.push("cond_string", "\\\\x[0-7]{H}", eHex, ".");
	rules.push("cond_string", "\\u{H}{4}|\\U{H}{6}", eUnicode, ".");
	rules.push("cond_string", "[^\\\n\"]", eChar, ".");
	rules.push("cond_string", "[\"]", TOK_STRING, "INITIAL");

	rules.push("[\n\r\t ]", rules.skip());

	rules.push("INITIAL", "[/][/]", rules.skip(), "cond_comment");
	rules.push("cond_comment", "\n", rules.skip(), "INITIAL");
	rules.push("cond_comment", "{UNICODE}", eError, ".");
	rules.push("cond_comment", "[^\n]", rules.skip(), ".");

	rules.push("INITIAL", "[/][*]", rules.skip(), "cond_comment");
	rules.push("cond_comment", "[*][/]", rules.skip(), "INITIAL");
	rules.push("cond_comment", "{UNICODE}", eError, ".");
	rules.push("cond_comment", ".{+}[\r\n]", rules.skip(), ".");

	rules.push("<=", LE);
	rules.push(">=", GE);
	rules.push("==", EQ);
	rules.push("!=", NE);
	rules.push("&&", AND);
	rules.push("[|][|]", OR);

	rules.push(".", eDot);
}
*/

void char8_test()
{
	using rules = lexertl::basic_rules<char, char, unsigned char>;
	using state_machine = lexertl::basic_state_machine<char, unsigned char>;
	using generator = lexertl::basic_generator<rules, state_machine>;
	using debug = lexertl::basic_debug<state_machine, char, unsigned char>;
	using cmatch = lexertl::match_results<const char *, unsigned char>;
	using citerator = lexertl::iterator<const char *, state_machine, cmatch>;

	rules r;
	state_machine sm;

	r.push("[]a[-]", 1);
	generator::build(r, sm);
	debug::dump(sm, std::cout);
	std::string input_("\"stdafx.h\"");

	citerator iter_(input_.c_str(), input_.c_str() + input_.size(), sm);
	citerator end_;

	for (; iter_ != end_; ++iter_)
	{
	}
}

void char16_test()
{
	lexertl::wrules rules;
	lexertl::wstate_machine sm;

	rules.push(L"[]a[-]+", 1);
	lexertl::wgenerator::build(rules, sm);
	lexertl::wdebug::dump(sm, std::wcout);
	std::wstring input_(L"\"stdafx.h\"");

	lexertl::wciterator iter_(input_.c_str(), input_.c_str() + input_.size(), sm);
	lexertl::wciterator end_;

	for (; iter_ != end_; ++iter_)
	{
	}
}

void char32_test()
{
	lexertl::u32rules rules;
	lexertl::u32state_machine sm;

	rules.push(U"[]a[-]", 1);
	lexertl::u32generator::build(rules, sm);
	lexertl::basic_debug<lexertl::u32state_machine, wchar_t>::dump(sm, std::wcout);
	std::u32string input_(U"\"stdafx.h\"");

	lexertl::u32citerator iter_(input_.c_str(), input_.c_str() + input_.size(), sm);
	lexertl::u32citerator end_;

	for (; iter_ != end_; ++iter_)
	{
	}
}

void case_mapping()
{
	lexertl::rules rules;
	lexertl::state_machine sm;
	lexertl::memory_file mf_("examples/unicode/UnicodeData.txt");
	lexertl::cmatch results_(mf_.data(), mf_.data() + mf_.size());
	enum e_Token { eEOF, eCodeValue, eName, eLl, eLu, eNeither, eMapping, eEmpty };
	e_Token eToken = eEOF;
	std::string code_;
	std::string mapping_;
	int count_ = 0;
	std::map<std::size_t, std::size_t> map;

	rules.push_state("NAME");
	rules.push_state("TYPE");
	rules.push_state("Ll");
	rules.push_state("Lu");
	rules.push_state("MAPPING");
	rules.push_state("END");

	rules.push("INITIAL", "^[0-9A-F]{4,6};", eCodeValue, "NAME");
	rules.push("NAME", "[^;]*;", sm.skip(), "TYPE");
	rules.push("TYPE", "Ll;", eLl, "Ll");
	rules.push("Ll", "([^;]*;){9}", sm.skip(), "MAPPING");
	rules.push("TYPE", "Lu;", eLu, "Lu");
	rules.push("Lu", "([^;]*;){10}", sm.skip(), "MAPPING");
	rules.push("TYPE", "[^;]*;", eNeither, "END");
	rules.push("MAPPING", ";", eEmpty, "END");
	rules.push("MAPPING", "[0-9A-F]{4,6};", eMapping, "END");
	rules.push("END", "[^\n]*\n", sm.skip(), "INITIAL");
	lexertl::generator::build(rules, sm);

	do
	{
		lexertl::lookup(sm, results_);
		eToken = static_cast<e_Token>(results_.id);

		if (eToken == eEOF)
		{
			break;
		}
		else if (eToken != eCodeValue)
		{
			throw std::runtime_error("Syntax error");
		}

		code_ = results_.str();
		lexertl::lookup(sm, results_);
		eToken = static_cast<e_Token>(results_.id);

		if (eToken != eLl && eToken != eLu && eToken != eNeither)
		{
			throw std::runtime_error("Syntax error");
		}

		if (eToken != eNeither)
		{
			lexertl::lookup(sm, results_);
			eToken = static_cast<e_Token>(results_.id);

			if (eToken == eMapping)
			{
				std::stringstream ss_;
				std::size_t code_val_;
				std::size_t mapping_val_;

				mapping_ = results_.str();
				ss_ << code_;
				ss_ >> std::hex >> code_val_;
				ss_.str("");
				ss_ << mapping_;
				ss_ >> std::hex >> mapping_val_;
				map[code_val_] = mapping_val_;

//				std::cout << "(0x" << code_.substr(0, code_.size() - 1) << ", " <<
//					"0x" << mapping_.substr(0, mapping_.size() - 1) << "), ";
				code_.clear();
				mapping_.clear();
				++count_;
			}
		}
	} while (results_.id != 0);

	std::pair<std::size_t, std::size_t> first_ = { 0, 0 };
	std::pair<std::size_t, std::size_t> second_ = { 0, 0 };

	for (auto &pair : map)
	{
		if (first_.first == 0)
		{
			first_.first = first_.second = pair.first;
			second_.first = second_.second = pair.second;
		}
		else if (pair.first == first_.second + 1 &&
			(pair.second == second_.second + 1 ||
			pair.second <= pair.first &&
			pair.second == second_.second - 1))
		{
			first_.second = pair.first;
			second_.second = pair.second;
		}
		else
		{
			std::cout << "            {{0x" << std::hex << std::setfill('0') << std::setw(4) << first_.first <<
				", 0x" << std::setfill('0') << std::setw(4) << first_.second <<
				"}, {0x" << std::setfill('0') << std::setw(4) << second_.first << ", 0x" <<
				std::setfill('0') << std::setw(4) << second_.second << "}},\n";
			first_.first = first_.second = pair.first;
			second_.first = second_.second = pair.second;
		}
	}

	std::cout << "            {{0x" << std::hex << std::setfill('0') << std::setw(4) << first_.first <<
		", 0x" << std::setfill('0') << std::setw(4) << first_.second <<
		"}, {0x" << std::setfill('0') << std::setw(4) << second_.first << ", 0x" <<
		std::setfill('0') << std::setw(4) << second_.second << "}},\n";
}

void regex_tokeniser()
{
	enum eToken
	{
		eEOF, eRegEx
	};
	lexertl::rules rules;
	lexertl::state_machine sm;

	rules.insert_macro("posix_name", "alnum|alpha|blank|cntrl|digit|graph|"
		"lower|print|punct|space|upper|xdigit");
	rules.insert_macro("posix", "\\[:{posix_name}:\\]");
	rules.insert_macro("escape", "\\\\([^xuc]|x\\d{2}|u\\d{4}|c[a-zA-Z])");

	rules.push("{escape}|"
		"(\\[^?({escape}|{posix}|[^\\\\\\]])*\\])|"
		"\\^|"
		"\\$|"
		"[|]|"
		"([*+?][?]?)|"
		"([{][0-9]+(,([0-9]+)?)?[}])|"
		"([(]([?][-risx]*[:])?)|"
		"[)]|"
		"[\\x21-\\xff]", eRegEx);

	lexertl::generator::build(rules, sm);

	std::string input = "[/][*](.|\\n)*?[*][/]";
	lexertl::citerator iter(input.c_str(), input.c_str() + input.size(), sm);
	lexertl::citerator end;

	for (; iter != end; ++iter)
	{
		std::cout << iter->str() << ", " << iter->id << '\n';
	}
}

#include <fstream>
#include "include/lexertl/stream_shared_iterator.hpp"

struct test
{
	const char *regex;
	const char *text;
	std::size_t start;
	std::size_t end;
};

void test_ng_repeats()
{
	test tests[] =
	{
/*		{"a*?", "aa", 0, 0},
		{"^a*?$", "aa", 0, 2},
		{"^.*?$", "aa", 0, 2},
		{"^(a)*?$", "aa", 0, 2},
		{"^[ab]*?$", "aa", 0, 2},
		{"a??", "aa", 0, 0},
		{"a+?", "aa", 0, 1},
		{"a{1,3}?", "aaa", 0, 1},
		{"\\w+?w", "...ccccccwcccccw", 3, 11},
		{"\\W+\\w+?w", "...ccccccwcccccw", 0, 11},*/
		// This generates the wrong state machine
		{"abc|\\w+?", "abd", 0, 1},
/*		{"abc|\\w+?", "abcd", 0, 3},
		{"<\\s*tag[^>]*>(.*?)<\\s*[/]tag\\s*>",
		" <tag>here is some text<[/]tag> <tag><[/]tag>", 0, 30},
		{"<\\s*tag[^>]*>(.*?)<\\s*[/]tag\\s*>",
		" < tag attr=\"something\">here is some text< [/]tag > <tag><[/]tag>", 0, 52},
		{"xx-{0,2}?(?:[+-][0-9])??\\z", "xx--", 0, 5},
		{"xx.{0,2}?(?:[+-][0-9])??\\z", "xx--", 0, 5},
		{"xx.{0,2}?(?:[+-][0-9])??\\z", "xx--", 0, 5},
		{"xx[/-]{0,2}?(?:[+-][0-9])??\\z", "xx--", 0, 5}*/
	};

	for (const auto &t : tests)
	{
		lexertl::rules rules;
		lexertl::state_machine sm;

		rules.push(t.regex, 1);
		lexertl::generator::build(rules, sm);

		lexertl::citerator iter(t.text, t.text + strlen(t.text), sm);

		std::cout << iter->str() << ", " << iter->id << '\n';
	}
}

int main()
{
	try
	{
		lexertl::rules rules;
		//lexertl::state_machine sm;

		/*
		rules.push("a$", 1);
		rules.push("a\nb", 2);
		lexertl::generator::build(rules, sm);
		lexertl::debug::dump(sm, std::cout);
		*/

/*
		using string_token = lexertl::basic_string_token<char>;
		string_token token;

		token.insert(string_token::range('\0', '\xff'));
		token.remove(string_token('g', 'g'));
		token.negate();

		int i = 0;
*/
/*
		std::string input = "abc xyz";
		lexertl::rules rules;
		lexertl::state_machine sm;

		// this does NOT match the first string
		rules.insert_macro("ALPHA", "^[A-Za-z]+");
		rules.push("{ALPHA}", 1);
		lexertl::generator::build(rules, sm);
		lexertl::smatch results(input.begin(), input.end());

		do
		{
			lexertl::lookup(sm, results);
			int rid = results.id;

			if (!rid) break;

			std::cout << "Id: " << rid << ", Token: '" <<
				results.str() << "'\n";
		} while (true);
*/
		//case_mapping();
		//lexertl::rules rules(lexertl::match_zero_len | lexertl::dot_not_cr_lf);
		lexertl::char_state_machine sm;

		//rules.push("a*", 1);
		//rules.push("a*?", 1);
		//rules.push("[/][*].{+}[\r\n]*[*][/]", 1);
		//rules.push("[/][*].{+}[\r\n]*?[*][/]", 1);
		//rules.push("d.{+}[\r\n]*s[^;]*", 1);
		//rules.push("d.{+}[\r\n]*?s[^;]*", 1);
		//rules.push("a*a", 1);
		//rules.push("a*?a", 1);
		//rules.push("a*b", 1);
		//rules.push("a*?b", 1);
		//rules.push("a*[^;]*", 1);
		//rules.push("a*?[^;]*", 1);
		//rules.push("\\w??|a", 1);
		//rules.push("a|\\w??", 1);
		//rules.push("abc|\\w+", 1);
		//rules.push("abc|\\w+?", 1);
		//rules.push("\\w+|abc", 1);
		//rules.push("\\w+?|abc", 1);
		//rules.push("(?s:.)*?b", 1);
		//rules.push("(?s:.)*?s[^;]*", 1);

		// Is it possible to change operators during tokenisation
		// to modify non-greedy operators?

		rules.push("abc|\\w+?", 1);
		rules.push("a{3}", 1);
		lexertl::basic_generator<lexertl::rules, lexertl::char_state_machine>::build(rules, sm);
		//lexertl::debug::dump(sm, std::cout);

		//test_ng_repeats();
/*
		rules.push("a?a?a?b", 1);
		rules.push("[a-z]c", 2);
		lexertl::generator::build(rules, sm);
		sm.minimise();
		lexertl::debug::dump(sm, std::cout);
*/

/*
		std::string input = "print";
		lexertl::u32rules rules;
		lexertl::u32char_state_machine sm;

		rules.push(U"print", 5);
		rules.push(U"[a-z]+", 4);
		rules.push(U"[0-9]+", 3);
		rules.push(U"0x[0-9a-f]+", 2);
		rules.push(U"[\\000-\\377]", 1);
		lexertl::u32char_generator::build(rules, sm);

		// char_state_machine has transitions ordered by state number.
		// This means that you cannot directly binary chop on the transitions.

		int i = 0;
*/
/*
lexertl::rules rules;
		lexertl::state_machine sm;
		std::ifstream ifi("C:\\Users\\Ben Hanson\\Desktop\\ssiterator.dat");
		lexertl::stream_shared_iterator iter;
		lexertl::stream_shared_iterator end;
		lexertl::match_results<lexertl::stream_shared_iterator> results;

		iter = lexertl::stream_shared_iterator(ifi);
		end = lexertl::stream_shared_iterator();
		results.reset(iter, end);
//		results = lexertl::match_results<lexertl::stream_shared_iterator>(iter, end);

		rules.push("[0-9]+", 1);
		rules.push("[a-z]+", 2);
		rules.push(".{+}[\r\n]", rules.skip());
		lexertl::generator::build(rules, sm);
		// Look ahead
		lexertl::lookup(sm, results);

		while (results.id != 0)
		{
			std::cout << "Id: " << results.id << ", Token: '" <<
				results.str() << "'\n";
			lexertl::lookup(sm, results);
		}
*/
/*
		rules.insert_macro("AllValid", "[\\x1-\\xff]");
		rules.insert_macro("ControlCodes", "[\\0-\\x1f\\x7f-\\x9f]");
		rules.insert_macro("AnyCh", "{AllValid}{-}{ControlCodes}");
		rules.push("{AnyCh}+", 123);

		lexertl::generator::build(rules, sm);
		lexertl::debug::dump(sm, std::cout);
*/

//		regex_tokeniser();
//		char8_test();
//		static_assert(!std::is_same<wchar_t, uint32_t>(), "Bang");

//		rules.push("[/][*].{+}[\r\n]*?[*][/]", 1);
//		rules.push("[/][*]([^*]|[*]+[^*/])*[*]+[/]", 1);
/*
		rules.push("ab", 1);
		rules.push("[a-z]+", 2);
		lexertl::generator::build(rules, sm);
		sm.minimise();
		lexertl::debug::dump(sm, std::cout);

		int i = 0;
*/
//		lexertl::table_based_cpp::generate_cpp("lookup", sm, true, std::cout);
//		lexertl::u32rules rules(lexertl::icase);

//		rules.push(U"[\\x100-\\x185]", 1);

//		case_mapping();
//		openscad();
//		char8_test();
//		char16_test();
//		char32_test();
/*
		std::regex rx("(int|short|long)|([a-zA-Z]+)|([0-9]+)|(=)", std::regex::extended);
		std::string s = "long longa=5;";
		std::sregex_iterator iter(s.begin(), s.end(), rx);
		std::sregex_iterator end;

		for (; iter != end; ++iter)
		{
			if ((*iter)[1].first != (*iter)[1].second)
			{
				std::cout << '\'' << (*iter)[1].str() << "': Keyword\n";
			}
			else if ((*iter)[2].first != (*iter)[2].second)
			{
				std::cout << '\'' << (*iter)[2].str() << "': Identifier\n";
			}
			else if ((*iter)[3].first != (*iter)[3].second)
			{
				std::cout << '\'' << (*iter)[3].str() << "': Number\n";
			}
			else if ((*iter)[4].first != (*iter)[4].second)
			{
				std::cout << '\'' << (*iter)[4].str() << "': Equals\n";
			}
		}
*/
//		rules.reverse();
//		static_assert(std::is_move_assignable<lexertl::detail::basic_re_token<char, char>>::value, "No");
//		rules.push("[/][*][^;]*?[*][/]", 1);
//		rules.push("[/][*].{+}[\r\n]*?[*][/]", rules.skip());
	}
	catch (const std::exception &e)
	{
		std::cout << e.what() << '\n';
	}

	return 0;
}
